///////////////////////////////////////////////////////////////////////////////////////////
// Purpose: Provide a viewer for viewing various tracks
// Note: 1. This class does not do alignments!!! Everything should be aligned!!!
//       2. Read function comments before calling them.
//       3. Provide a data provider to the viewer providing data in the format needed
// How to use: 1. Include Ext.
//             2. Include this class and its css.
//             3. Create a new instance 'var mv = BMS.AG.ML.DynamicViewer();
//             4. Make sure there's a data producer feeding data to this class
// Author: Mingyi Liu

// cookieConfig = { expires:, days:, cookie: } where cookie is new Ext.state.CookieProvider()
// default cookie (cookieConfig=null or {}) never expires (set expires:null to have cookie deleted when browser closes)
// for config info check _loadState()

BMS.AG.ML.DynamicViewer = function(cookieConfig, appConfig)
{
  this.stateManager = new BMS.AG.ML.DynamicViewer.stateManager(cookieConfig, this);
  var defConf =
    {
      id:'tablemap',
      className:'tablemap', // if changing className here, change the corresponding name in CSS file!
      sizeRange: {min:4, max:12},
      boundingDiv: { id:'divmap' },
      xOverlap: 9, // how much "Left", "Right" screens should overlap (in base units) and how much left/right margin for mutations viewed by "Find the Next Unseen Mutation" should have
      yOverlap: 2 // how much "Up", "Down" screens should overlap (in base units) and how much up/down margin for mutations viewed by "Find the Next Unseen Mutation" should have
    };
  Ext.apply(this, appConfig, defConf);
  if(appConfig && appConfig.controller)
  {
    this.viewController = new BMS.AG.ML.ViewerController(
      {
        callback: this._controllerCallback,
        scope: this,
        title: appConfig.controller.title,
        renderTo: appConfig.controller.renderTo
      });
  }
  this.initState();
};
BMS.AG.ML.DynamicViewer.prototype =
{
  //////////////////////////////////////////////////////////////////////
  // public functions
  setCookieKey: function(key)
  {
    this.stateManager.setKey(key);
  },
  // provide string description of where the current view window is at, useful for logs
  viewWindowString: function()
  {
    var conf = this.stateManager.config, tracks = [];
    if(conf.displayTracks.dna.show) tracks.push('dna');
    if(conf.displayTracks.prot.show) tracks.push('prot');
    var t = this.trackManager.get1stDisplayedTrack();
    return '(' + conf.databox.xstart + ' on ' + tracks[0] + 'track;' + (t? 'Sample:' + t.name + '-' + t.tclass : '') + '), displaying ' + tracks.join(',') + ' track' + (tracks.length > 1? 's' : '');
  },
  // move the crosshair (including viewController's)
  moveCross: function(e)
  {
    if(!this.stateManager.config) return; // nothing initiated yet
    var xy = e.getXY();
    if(this.stateManager.appConfig.useCross)
    {
      if(this.vl) this.vl.setX(xy[0] + 1);
      if(this.hl) this.hl.setY(xy[1] + 1);
    }
    if(this.controller && this.divmap)
    {
      var x = (xy[0] - this.divmap.getX() - this.idcolWidth) / (this.divWidth - this.idcolWidth);
      var z = this.stateManager.config.size * this.trackManager.countDisplayedTracks(1);
      var y = (xy[1] - this.divmap.getY() - z) / (this.divHeight - z);
      if(x < 0) x = 0;
      if(y < 0) y = 0;
      this.viewController.moveCross(x * this.stateManager.config.databox.width + this.stateManager.config.databox.xstart,
        y * this.stateManager.config.databox.height + this.stateManager.config.databox.ystart);
    }
  },
  // toggle the highlight of the 2 cells at the first row/col of currently hovered cell
  // highlight style is defined by td.hilite class and td class
  toggleCellHiLite: function(e)
  {
    if(!this.stateManager.config) return; // nothing initiated yet
    var info = this._getTargetInfo(e);
    if(!info.isCell) return;
    if(!this.stateManager.appConfig.useCross)
    {
      if(info.baseCellIndex > 0)
      {
        if(info.td.featureData && info.baseRowIndex >= 0)
          BMS.AG.ML.Util.toggleClass(info.rows[info.baseRowIndex].children[info.baseCellIndex], 'hilite');
        BMS.AG.ML.Util.toggleClass(info.rows[info.rowIndex].children[0], 'hilite');
      }
    }
    if(e.type == 'mouseover')
    {
      if(info.td.toolTip)
      {
        if(this.stateManager.appConfig.clickToPopup) info.td.toolTip.showDelay = 0; // reset showDelay based on current user setting
        else
        {
          info.td.toolTip.shouldShow = true;
          info.td.toolTip.showDelay = 500; // reset showDelay based on current user setting
          info.td.toolTip.show();
        }
      }
      else if(!this.stateManager.appConfig.clickToPopup)
        this._createToolTips(info.td, e); // lazy render tooltip now (BTW e.getTarget() is not good because idcol e.getTarget is a div, not a td)
    }
  },
  // turn on/off crosshair feature
  toggleCross: function()
  {
    if(!this.stateManager.config) return; // nothing initiated yet
    this.stateManager.appConfig.useCross = !this.stateManager.appConfig.useCross;
    var msg;
    if(!this.stateManager.appConfig.useCross)
    {
      if(this.vl) this.vl.setLeft(-10);
      if(this.hl) this.hl.setTop(-10);
      msg = 'Crosshair removed';
    }
    BMS.AG.ML.Util.alert('Info', msg || 'Crosshair activated');
    this.log('Option', msg? 'Remove Crosshair' : 'Show Crosshair');
    this.stateManager.recordState();
  },
  // turn on/off mouseover for popup
  togglePopup: function()
  {
    if(!this.stateManager.config) return; // nothing initiated yet
    this.stateManager.appConfig.clickToPopup = !this.stateManager.appConfig.clickToPopup;
    var msg;
    if(!this.stateManager.appConfig.clickToPopup)
      msg = 'Mouseover a feature/mutation to see details!';
    BMS.AG.ML.Util.alert('Info', msg || 'Click on a feature/mutation to see details!');
    this.log('Option', msg? 'Mouseover for popup' : 'Click for popup');
    this.stateManager.recordState();
  },
  toggleSeparators: function()
  {
    if(!this.stateManager.config) return; // nothing initiated yet
    this.stateManager.appConfig.showSep = !this.stateManager.appConfig.showSep;
    var msg;
    msg = this.stateManager.appConfig.showSep? 'Show group separators' : 'Do not show group separators';
    BMS.AG.ML.Util.alert('Info', msg);
    this.log('Option', msg);
    this.stateManager.recordState();
    this.refreshView();
  },
  refreshView: function()
  {
    if(this.data) this._drawMaskTracks();
  },
  // move the view to the specified position on the given track
  moveViewPosition: function(track, pos)
  {
    if(!this.stateManager.config) return; // nothing initiated yet
    var d = this.stateManager.config.databox;
    var base = this.trackManager.getBaseForClass(track);
    var xy = pos.split(',');
    var pos1 = base.addGap(xy[0]-0);
    if(track == 'prot' && this.hasDNATrack && this.stateManager.config.displayTracks.dna.show)
      d.xstart = pos1 * 3 - 2;
    else d.xstart = pos1;
    if(xy.length > 1)
    {
      if(isNaN(xy[1])) d.ystart = this.trackManager.getTrackNumberByName(xy[1], track);
      else d.ystart = xy[1]-0;
    }
    this.getDataDisplay();
    this.log('Action', 'Moved view to ' + track + ' track');
  },
  // move the view based on specified direction
  moveView: function(direction)
  {
    if(!this.stateManager.config) return; // nothing initiated yet
    var d = this.stateManager.config.databox;
    var x = d.width - this.xOverlap;
    var y = d.height - this.yOverlap;
    if(direction == 'Right')
    {
      if(0) // this one's most troublesome, will work on it later
        { BMS.AG.ML.Util.alert('Info','Already at rightmost!'); return; }
      d.xstart += x;
    }
    else if(direction == 'Down')
    {
      if(this.data && d.ystart + d.height > this.data.length)
        { BMS.AG.ML.Util.alert('Info','Already at bottom!'); return; }
      d.ystart += y;
    }
    else if(direction == 'Left')
    {
      if(d.xstart <= 1) { BMS.AG.ML.Util.alert('Info','Already at leftmost!'); return; }
      d.xstart -= (d.xstart > x)? x: d.xstart - 1;
    }
    else if(direction == 'Up')
    {
      if(d.ystart <= 1) { BMS.AG.ML.Util.alert('Info','Already at top!'); return; }
      d.ystart -= (d.ystart > y)? y: d.ystart - 1;
    }
    this.getDataDisplay();
    this.log('Action', 'Moved view ' + direction);
  },
  // zoom the view based on specified direction ('In', 'Out', or a number b/w 3-10)
  zoomView: function(direction)
  {
    if(!this.stateManager.config) return; // nothing initiated yet
    if(direction == 'In')
    {
      if(this.stateManager.config.size < this.sizeRange.max) this.stateManager.config.size += 5;
      else { BMS.AG.ML.Util.alert('Info','Already at biggest font!'); return; }
    }
    else if(this.stateManager.config.size > this.sizeRange.min) this.stateManager.config.size -= 5;
    else { BMS.AG.ML.Util.alert('Info','Already at maximum resolution!'); return; }
    this.idcolWidth = Math.floor(this.idcolRatio * this.stateManager.config.size);
    this.getDataDisplay();
    this.log('Action', 'Zoomed view ' + direction);
  },
  showTrack: function(trackName, show)
  {
    // make sure the view does not shift horizontally due to removing/showing tracks (this could happen when both dna and prot are displayed but then dna track gets unchecked, or prot is displayed, then dna gets checked)
    if(trackName == 'dna') // dna check/uncheck is what causes trouble
    {
      var widthlcf = this.trackManager.getWidthLcf('dna');
      if(widthlcf > 1 && show)
        this.stateManager.config.databox.xstart = this.stateManager.config.databox.xstart * 3 - 2;
        // !!!!!!!!!!!!! need to take offset into account
      else if(widthlcf == 1 && !show) // dna just got unchecked
        this.stateManager.config.databox.xstart = Math.floor((this.stateManager.config.databox.xstart - 1) / 3) + 1;
    }
    // the following code has to come after the code above, otherwise it makes the code above wrong
    this.stateManager.config.displayTracks[trackName].show = show;
    if(this.stateManager.config.displayTracks[trackName+'coords'])
      this.stateManager.config.displayTracks[trackName+'coords'].show = show;
    // make sure the view does not shift vertically either!
    // first find the current top track (ystart works in the space of ALL TRACKS THAT SHOULD be shown. Namely, if there are 100 nonbase tracks, 50 DNA-protein track pairs, and currently only protein track is shown. If ystart = 5, it means that protein track #5 all the way till ystart + height -1 will be displayed)
    this.stateManager.config.databox.ystart = this.trackManager.get1stDisplayedTrackNumber();
    this.stateManager.config.newControllerData = true; // meaning if we use view controller, the data for the controller needs update now
    this.getDataDisplay();
    this.log('Action', (show? 'Show ':'Hide ') + trackName + ' track');
  },
  filterComment: function(opt, noFilterFeatures)
  {
    if(!this.stateManager.config) return; // nothing initiated yet
    this.filter = opt;
    this._resetFilters();
    this._preFilter();
    this.stateManager.config.newControllerData = true; // meaning if we use view controller, the data for the controller needs update now
    this.filterApplied = true;
    this.stateManager.config.databox.ystart = 1; // move to top so that users don't get surprised when they don't see results in view
    this._drawMaskTracks(noFilterFeatures);
    var msg = 'Found ' + this.filterFound + ((this.filterFound > 1)?' match(es)!':' match!');
    this.log('Result', '  ' + msg);
  },
  filterFeature: function(opt)
  {
    if(!this.stateManager.config) return; // nothing initiated yet
    this.featureFilter = opt;
    this._filterFeatures(true);
  },
  resetFilters: function()
  {
    if(!this.stateManager.config) return; // nothing initiated yet
    if(this.filter || this.baseFilter || this.featureFilter)
    {
      this._resetFilters();
      this.stateManager.config.newControllerData = true; // meaning if we use view controller, the data for the controller needs update now
      delete this.featureFilter;
      this._drawMaskTracks();
      this.filterApplied = false;
      delete this.filter;
      delete this.baseFilter;
    }
  },
  // loads all data through ajax and displays
  getDataDisplay: function(options)
  {
    if(this.data) this._drawMaskTracks();
    else
    {
      var e = Ext.get(this.boundingDiv.id);
      if(!options) options = {};
      if(this.dataparam) Ext.apply(options, this.dataparam);
//       e.load({
//         url: this.dataurl,
//         callback:this._drawTracksCallback,
//         scope:this,
//         params: options,
//         text: 'Contacting Server to retrieve data ...'
//       });
      e.mask('Contacting Server to retrieve data ...', 'loading-indicator');
      Ext.Ajax.request({
        url: this.dataurl,
        scope:this,
        params: options,
        callback:this._drawTracksCallback
      });
    }
  },
  clearToolTips: function(e)
  {
    var info = this._getTargetInfo(e);
    if(info.isCell && info.td.toolTipInfo && this.stateManager.appConfig.clickToPopup) // clicking on a feature
      this._createToolTips(info.td, e, true); // lazy render tooltip now (BTW e.getTarget() is not good because idcol e.getTarget is a div, not a td)
    else if(this.toolTips)
      for(var i = 0; i < this.toolTips.length; i++)
        this.toolTips[i].hide();
    if(this.editWins)
      for(var i = 0; i < this.editWins.length; i++)
        this.editWins[i].close();
  },
  updateComment: function(e, info) // event by 'afteredit'
  {
    this.commentConf.updateComment(e, info, this.comments, this.trackManager.getTrack(info.track.id).comments, this.currentUser); // delegate to the obj that knows comment structure
  },
  removeComment: function(e, info)
  {
    this.commentConf.removeComment(e, info, this.comments, this.trackManager.getTrack(info.track.id).comments); // delegate to the obj that knows comment structure
  },
  removeAll: function()
  {
    delete this.data;
    delete this.stateManager.config;
    delete this.stateManager.appConfig; // had to delete this due to Ext.apply bug
    delete this.comments;
    delete this.controllerData;
    // destroy all Ext obj. (deal with comment/grid later...)
    this._destroyToolTips();
  },
  initState: function()
  {
    this.trackManager = new BMS.AG.ML.DynamicViewer.TrackManager(null, this);
    this.stateManager.loadState(); // load defaults or previous config stored in cookie
    var oldWidth = BMS.AG.ML.Util.getCSSRule('.' + this.className + ' td.idcol', 'width', {isNum:true});
    var oldSize = BMS.AG.ML.Util.getCSSRule('.' + this.className + ' td', 'width', {isNum:true});
    this.idcolRatio = oldWidth / oldSize; // record original CSSrules-intended ratio (this way guarantees consistent width after cookie stores different size from CSS-defined yet idcol not recorded in cookie (which is another solution))
    this.idcolWidth = Math.floor(this.idcolRatio * this.stateManager.config.size);
  },
  filterBases: function(type, position, bases)
  {
    var tracks = this.trackManager.filterBases(type, position, bases);
    this.stateManager.config.newControllerData = true;
    this._drawMaskTracks();
    var all = new Array(), msg = 'Found ' + tracks.length + ((tracks.length > 1)?' match(es)!':' match!');
//     BMS.AG.ML.Util.alert('Info', msg);
    for(var i in bases) all.push(i);
    this.log('Action', 'Filtered ' + type + ' track for ' + all.join(',') + ' at position:' + position);
    this.log('Result', '  ' + msg);
    this.baseFilter = true;
  },
  log: function(type, msg)
  {
    if(this.logger) this.logger.add(type, msg);
  },
  resetMapSize: function()
  {
    if(this.viewController.initiated)
    {
      var oldWidth = this.stateManager.config.databox.width;
      var oldHeight = this.stateManager.config.databox.height;
      this._setBoxSize();
      if(oldWidth < this.stateManager.config.databox.width || oldHeight < this.stateManager.config.databox.height)
        this._drawMaskTracks();
      else
        this.viewController.updateRect(this._getCurrentRect());
    }
    // if not initiated the _drawTracks will update the rect anyhow after drawing controller
  },
  setMask: function(msg)
  {
    if(!this.mask) this.mask = new Ext.LoadMask(this.boundingDiv.id, {msg:msg || 'Loading data...'});
    else if(!msg && this.maskOldMsg) this.mask.msg = this.maskOldMsg;
    else if(msg) { this.maskOldMsg = this.mask.msg; this.mask.msg = msg; }
  },
  /////////////////////////////////////////////////////////////////////////
  // private functions
  _drawMaskTracks: function(noFilterFeatures)
  {
    // before drawing anything, let's be user friendly and show 'loading...'
    this.setMask();
    this.mask.show();
    // I hate to use a timeout to delay calling so that mask has time to show
    var self = this;
    setTimeout(function(){self._drawTracks(noFilterFeatures)}, 10); // no need to use Function.call here
  },
  _setCellSize: function()
  {
    BMS.AG.ML.Util.updateCSSRule('.' + this.className + ' td','width',this.stateManager.config.size + 'px');
    BMS.AG.ML.Util.updateCSSRule('.' + this.className + ' td','min-width',this.stateManager.config.size + 'px');
    BMS.AG.ML.Util.updateCSSRule('.' + this.className + ' td','height',this.stateManager.config.size + 'px');
    BMS.AG.ML.Util.updateCSSRule('.' + this.className + ' td','line-height',this.stateManager.config.size + 'px');
    BMS.AG.ML.Util.updateCSSRule('.' + this.className + ' td','font-size',(this.stateManager.config.size/16) + 'em');
    // we have to update both as one's for IE, one's for FF
    BMS.AG.ML.Util.updateCSSRule('.' + this.className + ' td.idcol div.idcol', 'width', this.idcolWidth + 'px');
    BMS.AG.ML.Util.updateCSSRule('.' + this.className + ' td.idcol', 'width', this.idcolWidth + 'px');
  },
//   _drawTracksCallback: function(div,success,result)
  _drawTracksCallback: function(opts,success,result)
  {
    if(!success)
    {
      BMS.AG.ML.Util.alert('Warning','Cannot retrieve data from server - server/service might be down?');
      return;
    }
//     div.dom.innerHTML = '';
//     console.time('parsejson');
    var data = Ext.decode(result.responseText);
//     console.timeEnd('parsejson');
    if(data && data.seqdata)
    {
      // first check if dataurl/dataparam match what were stored in cookie or not, if not, resetState
      if(this.dataurl != this.stateManager.config.dataurl ||
         !BMS.AG.ML.Util.sameObj(this.dataparam, this.stateManager.config.dataparam))
      {
        this.stateManager.clearState();
        this.stateManager.loadState();
      }
      // now check and prevent a rarely-happening problem where both DNA and protein tracks got turned off in one user's browser every few months
      var tc = this.stateManager.config.displayTracks;
      if(tc)
      {
        if(tc.dna && !tc.dna.show && !tc.prot.show)
        {
          tc.dna.show = true;
          tc.prot.show = true;
        }
        if(tc.dna && !tc.dnacoords.show && !tc.protcoords.show)
        {
          tc.dnacoords.show = true;
          tc.protcoords.show = true;
        }
      }
      this.data = data.seqdata;
      this._insertCoordsTracks(); // insider coordinates tracks
//       this.currentUser = data.extra && data.extra.context ? data.extra.context : 'Myself';
      this.timestamp = (data.extra && data.extra.time)? data.extra.time : -1;
      this.trackManager.addTracks(this.data);
      this.comments = data.comment || {};
      if(!this.comments.row) this.comments.row = {};
      if(!this.comments.col) this.comments.col = {};
      if(!this.comments.cell) this.comments.cell = {};
      this._orgComByRow();
      this.stateManager.config.newControllerData = true;
      this._drawMaskTracks();
    }
  },
  _insertCoordsTracks: function()
  {
    var newdata = new Array(), id = -1;
    this.step = 5;
    for(var i = 0; i < this.data.length; i++)
    {
      var t = this.data[i];
      newdata.push(t);
      if(t.type == 'base' && t.tclass.match(/^dna|prot$/))
      {
        var tmp = { id:id--, name:t.tclass + ' coordinates', tclass:t.tclass + 'coords', type:'base', offset:t.offset, seq:[{ tx:1 }] };
        tmp.step = (t.tclass == 'dna')? 10 : 5;
        var newseq = new Array();
        var sarr = t.seq[0].s.split('');
        for(var j = 0, k = 0, l = 0; j < sarr.length; j++)
        {
          if(sarr[j] != ' ' && sarr[j] != '-') // special char
          {
            if((k+1) % tmp.step == 0)
            {
              newseq.push(k + 1);
              l = ((k + 1) + ' ').length - 1;
            }
            else if(--l <= 0) newseq.push(' ');
            k++;
          }
          else newseq.push(' ');
        }
        tmp.seq[0].s = newseq.join('');
        newdata.push(tmp);
      }
    }
    this.data = newdata;
  },
  // set comments as data[i].comments[tagName][context] = value array
  _setComment: function(c, d)
  {
    for(var i = 0; i < d.length; i++)
    {
      var com = d[i];
      if(!c[com[0]])
      {
        c[com[0]] = {};
        c[com[0]][com[2]] = [com[1]];
      }
      else
      {
        if(!c[com[0]][com[2]]) c[com[0]][com[2]] = new Array();
        c[com[0]][com[2]].push(com[1]);
      }
    }
  },
  _filterFeatures: function(isFilter) // isFilter == false -> reset feature filters
  {
    this.setMask('Filtering...');
    this.mask.show();
    var self = this;
    setTimeout(function(){self._filterF(isFilter)}, 10);
  },
  _filterF: function(isFilter)
  {
    var rows = this.map.select('tr'), tracks = this.trackManager.getDisplayedTracks();
    for(var i = 0; i < rows.getCount(); i++)
    {
      var track = tracks[i];
      var cells = rows.item(i).select('td');
      for(var j = 0; j < cells.getCount(); j++)
      {
        var cell = cells.item(j);
        if(cell.dom.featureData)
        {
          cell.setVisibilityMode(Ext.Element.VISIBILITY);
          var t = this.featureFilter.fn.call(this.featureFilter.scope, cell.dom.featureData, track);
          if(typeof(t) == 'object')
          {
            if(!isFilter)
            {
              if(cell.dom.oldInnerHTML) cell.dom.innerHTML = cell.dom.oldInnerHTML;
              cell.setVisible(true);
            }
            else if(t.length)
            {
              var max = (track.tclass == 'dna' || !this.stateManager.config.displayTracks['dna'].show)? 1 : 3; // use widthlcf is safer if there are more unitwidths than 'dna'/'prot' (1/3)
              cell.dom.oldInnerHTML = cell.dom.innerHTML;
              if(t.length > max) cell.dom.innerHTML = 'X';
              else cell.dom.innerHTML = t.join('');
            }
            else cell.setVisible(false);
          }
          else cell.setVisible(!isFilter || t);
        }
      }
    }
    // now let's deal with viewController
    if(this.controller)
      this.viewController.repaintFeatures(this._getFeatData(isFilter));
    this.mask.hide();
    this.setMask(); // reset msg
  },
  // organize all comments by row as filtering is row-based filtering
  _orgComByRow: function()
  {
    if(this.comments)
    {
      var row = this.comments.row, col = this.comments.col, cell = this.comments.cell;
      var data = {}, tracks = this.trackManager.tracks.all;
      for(var i = 0; i < tracks.length; i++)
      {
        var d = tracks[i];
        data[d.id] = d;
        if(!d.comments) d.comments = {};
        if(row[d.id]) this._setComment(d.comments, row[d.id]);
      }
      for(var id in cell)
      {
        var tmp = id.split('::');
        this._setComment(data[tmp[0]].comments, cell[id]);
      }
      for(var id in col)
      {
        var tmp = id.split('::');
        for(var i = 0; i < tracks.length; i++)
        {
          var d = tracks[i];
          if(d.tclass != data[tmp[0]].tclass) continue;
          this._setComment(d.comments, col[id]);
        }
      }
    }
  },
  _resetFilters: function()
  {
    var data = this.trackManager.tracks.nonbase;
    for(var i = 0; i < data.length; i++)
      data[i].failedFilter = false;
  },
  // we only filter comment for now
  _preFilter: function()
  {
    var data = this.trackManager.tracks.nonbase;
    this.filterFound = 0;
    for(var i = 0; i < data.length; i++)
    {
      var d = data[i];
      if(d.type != 'base')
      {
        if(!this.filter.fn.call(this.filter.scope, d.comments))
          d.failedFilter = true;
        else this.filterFound++;
      }
    }
  },
  // set the metrics of the current viewing window
  _setBoxSize: function()
  {
    this.divWidth = this.divmap.getWidth() - 2; // account for borders
    this.divHeight = this.divmap.getHeight() - 2;
    // calculate the width and height so that we fill up the div area entirely.
    this.stateManager.config.databox.width = Math.floor((this.divWidth - this.idcolWidth)/this.stateManager.config.size);
    this.stateManager.config.databox.height = Math.floor(this.divHeight / this.stateManager.config.size) - 1;
  },
  // the main function that draws the entire viewing window
  _drawTracks: function(noFilterFeatures)
  {
    var div = this.boundingDiv.id;
    var numtracks = 0, colTipsAdded = false;

    this._destroyToolTips();

    //////////////////////////////////////////////////////////////////////////////////////////////////////
    // This statement gets Ext.Element with table as a child, could be more useful/x-browser in the future
    this.divmap = Ext.get(div).update('<table id="' + this.id + '" class="' + this.className + '"></table><div id="vertline" class="vertline"></div><div id="horiline" class="horiline">&nbsp;</div>\n'); // &nbsp; needed in IE!
    this._setBoxSize();

    this.map = this.divmap.child('#'+this.id);
    this.vl = Ext.get("vertline");
    this.vl.setHeight(this.divHeight);
    this.hl = Ext.get("horiline");
    this.hl.setWidth(this.divWidth);

    this._setCellSize();

    // let's display the base tracks first
    var data = this.trackManager.getDisplayedTracks();

    // display track by track
    for(var i = 0; i < data.length; i++)
    {
      var t = data[i], b = data[i].base || t;
      var row = this.map.insertHtml('beforeEnd', t.getHtml()).children[0];

      // now put in any highlight data for the track
      this._setToolTip(row.children[0],
        {
          track: t, tgstart: b.tstart, tgend: b.tend, tstart: t.removeGap(b.tstart), tend: t.removeGap(b.tend), type: 'row', name:t.name
        });
      if(t.type == 'base' && t.tclass.match(/^(dna|prot)$/) && !colTipsAdded)
      {
        var skipped = 1;
        for(var m = skipped; m < row.children.length; m++)
        {
          if(row.children[m].innerHTML.match(/^[ -]?$/)) // special char
            skipped++;
          else
            this._setToolTip(row.children[m],
              {
                track: t,  tgstart: t.tstart + m - 1, tstart: t.removeGap(t.tstart) + m - skipped, type: 'col', seq: row.children[m].innerHTML, name:t.name
              });
        }
//         colTipsAdded = true; // this prevents adding protein col comments
      }
      else if(t.type != 'base' && t.seq && t.seq.length)
      {
        for(var m = 0; m < t.featureInView.length; m++)
        {
          var cellIndex = t.featureInView[m][1];
          row.children[cellIndex].featureData = t.featureInView[m][0];
          var x = t.featureInView[m][0].tx;
          this._setToolTip(row.children[cellIndex],
            {
              track: t, tgstart: x, tstart: t.removeGap(x), type: 'cell', seq: t.getBaseSeqAt(x, true), features: t.featureInView[m][0], name:t.name
            });
        }
      }
    }
    if(this.stateManager.appConfig.showSep)
    {
      var seppos = this.trackManager.getSepPositions();
      if(seppos.length)
      {
        for(var i = 0; i < seppos.length; i++)
        {
          var e = this.divmap.insertHtml('beforeEnd','<div class="groupsep">&nbsp;</div>',true);
          e.setWidth(this.divWidth);
          e.setY(this.divmap.getY() + this.stateManager.config.size * seppos[i]);
        }
      }
    }
    if(this.mask) this.mask.hide();
    this.stateManager.recordState();
    Ext.QuickTips.init();
    if(this.controller)
    {
      if(this.stateManager.config.newControllerData)
      {
        this.viewController.setData(this._getControllerData());
        this.viewController.updateView();
      }
      if(!this.calledByController)
        this.viewController.updateRect(this._getCurrentRect());
      this.calledByController = false;
      this.stateManager.config.newControllerData = false;
    }
    if(this.featureFilter && !noFilterFeatures) this._filterFeatures(true);
    if(this.finishedCallback) this.finishedCallback.fn.apply(this.finishedCallback.scope);
    this.log('View', 'Viewing ' + this.viewWindowString());
  },
  _getCurrentRect: function()
  {
    return [this.stateManager.config.databox.xstart, this.stateManager.config.databox.ystart, this.stateManager.config.databox.width, this.stateManager.config.databox.height];
  },
  _getMaxDataWidth: function()
  {
    var max = 0, data = this.trackManager.tracks.all;
    for(var i = 0; i < data.length; i++)
    {
      if(data[i].noDisplay()) continue;
      if(data[i].seq && data[i].seq.length)
      {
        for(var m = 0; m < data[i].seq.length; m++)
        {
          // we assume DNA longer than protein * 3 + offset, so just compare x
          var x = data[i].seq[m].x || data[i].seq[m].tx || 1;
          if(data[i].seq[m].s) x += data[i].seq[m].s.length - 1;
          if(x > max) max = x;
        }
      }
    }
    return max;
  },
  _getControllerData: function()
  {
    // for some reason I have to put the following on one line, otherwise get 'invalid label error'
    return { len: [this._getMaxDataWidth(), this.trackManager.countDisplayedTracks(2)], rect: this._getCurrentRect(), featOnly: true, color: {}, feat: this._getFeatData() };
  },
  _getFeatData: function(isRepaint) // format feature data for viewController (isRepaint indicates that get data for repaint purpose only!
  {
    var d = new Array(), data = this.trackManager.tracks.nonbase, track = 0;

    for(var i = 0; i < data.length; i++)
    {
      if(data[i].noDisplay()) continue;
      var offset = data[i].getOpt('offset');
      var unitwidth = data[i].getOpt('width');
      track++;
      if(data[i].seq && data[i].seq.length)
      {
        var seenx = {};
        for(var m = 0; m < data[i].seq.length; m++)
        {
          // let's just record the start position
          var x = data[i].seq[m].x || data[i].seq[m].tx || 1;
          // seenx: guard against features reported twice by accident; do not show nodata features in viewController
          if(seenx[x] || (data[i].seq[m].b && data[i].seq[m].b.length == 1 && data[i].seq[m].b[0][0].match(/[ .]/))) continue;
          seenx[x] = 1;
          var orig = data[i].seq[m].orig[0];
          var v = [(x - 1) * unitwidth + offset, track, 1];
          if(data[i].seq[m].s && !isRepaint) d.push([(x - 1) * unitwidth + offset, track, 1]);
          else if (data[i].seq[m].b && data[i].seq[m].b.length > 0)
          {
            var sum = 0, pushData = true;
            if(isRepaint)
            {
              var t = this.featureFilter.fn.call(this.featureFilter.scope, data[i].seq[m], data[i]);
              if(typeof(t) == 'object') pushData = (t.length > 0);
              else pushData = t;
            }
            if(pushData)
            {
              for(var n = 0; n < data[i].seq[m].b.length; n++)
                if(data[i].seq[m].b[n][0] != orig) sum += data[i].seq[m].b[n][1];
              d.push([(x - 1) * unitwidth + offset, track, sum / 100]);
            }
          }
        }
      }
    }
    if(!isRepaint) this.controllerData = d;
    return d;
  },
  _guessUnit: function(tclass)
  {
    switch(tclass)
    {
      case 'prot':
      case 'domain': return 'aa';
      default: return 'nt';
    }
  },
  _setToolTip: function(e, info)
  {
    if(!this.toolTips) this.toolTips = new Array();
    var u = this._guessUnit(info.track.tclass);
    var desc = (info.track.desc)? '<p>' + info.track.desc : '';
    var d = ((!info.tend || info.tstart == info.tend)? info.tstart : info.tstart + ' - ' + info.tend) + ' ' + u;
    var dg = ((!info.tgend || info.tgstart == info.tgend)? info.tgstart : info.tgstart + ' - ' + info.tgend) + ' ' + u; // gapped
    if(d != dg) d += ' or ' + dg + ' (gapped)';
    var h, ihtml; // infoHtml
    if(info.type == 'row')
    {
      h = info.track.name + '<p>' + d + desc;
      ihtml = info.track.infoHtml;
    }
    else if(info.type == 'col')
      h = info.track.name + '<p>' + info.seq.toUpperCase() + ' at ' + d + desc;
    else 
    {
      // now add mutation info
      var md = info.features;
      ihtml = md.infoHtml;
      if(md && md.b.length > 0)
      {
        md.b.sort(function(a,b){return a[1]-b[1]});
        var sum = 0, orig = info.seq, str;
        for(var n = 0; n < md.b.length; n++)
          if(md.b[n][0] != orig && md.b[n][0] != '.') sum += md.b[n][1]; // special char
        d += ', ' + orig + ', ' + (100 - sum) + '%';
        if(sum) // sum could be 0 if the mutation is '.'
        {
          d += '<br><p><b>Mutation Info:</b><br/>\n<table class="x-tip-body"><tr>';
          for(var n = 0; n < md.b.length; n++)
            if(md.b[n][0] != orig && md.b[n][0] != '.') // special char
              d += '<td>' + md.b[n][0] + '</td>';
          d += '</tr><tr>';
          for(var n = 0; n < md.b.length; n++)
            if(md.b[n][0] != orig && md.b[n][0] != '.') // special char
              d += '<td>' + md.b[n][1] + '%</td>';
          d += '</tr></table></p>\n';
        }
      }
      h = info.track.name + '<p>' + d;
    }
    var el = e.dom? e.dom : e;
    info.h = h;
    info.ihtml = ihtml;
    
    var c = this.comments;
    info.com = (info.type == 'row' && c.row)? c.row[info.track.id] : (info.type == 'col' && c.col)? c.col[info.track.id+'::'+info.tstart] : (c.cell)? c.cell[info.track.id+'::'+info.tstart] : null;
    // need to improve further so that when all user comments were deleted, lowlite class is removed.
    if(info.com  && info.com.length && (info.type != 'cell' || this._hasUserCom(info.com))) replaceColor(el, 'lowlite');
    el.toolTipInfo = info;
  },
  _hasUserCom: function(com)
  {
    if(com && com.length)
    {
      for(var i = 0; i < com.length; i++)
        if(com[i][0] == 'Comment') return true;
    }
    return false;
  },
  _hiliteTD: function(e)
  {
    if(!e.resizer)
    {
      e.resizer = new Ext.Resizable(e.id,
      {
        handles: 'all',
        minWidth: 300,
        minHeight: 200,
        pinned: true
      });
      e.resizer.on('resize', dynamicViewer._resizeChildren, dynamicViewer);
    }
    replaceColor(e.target.dom, 'hilite')
  },
  // improve later
  _resizeChildren: function(e, w, h, event)
  {
    var el = e.el, cmp = Ext.getCmp(el.id), t = Ext.getCmp(el.child('div.x-tab-panel').id);
    t.setWidth(w - 12);
    t.setHeight(h - 18);
    if(cmp.syncShadow) cmp.syncShadow();
//     var tip = Ext.getCmp(e.el.id);
//     tip.body.setWidth(w - 12);
//     tip.body.setHeight(h - 18);
//     tip.syncSize();
  },
  _loliteTD: function(e)
  {
    restoreColor(e.target.dom, 'hilite')
  },
  // this function is for lazy rendering
  _createToolTips: function(e, event, isClick)
  {
    var el = Ext.get(e);
    if(!e.toolTipInfo || (e.parentNode && /^(dna|prot)coords$/.test(e.parentNode.className))) return;
    if(e.toolTip)
    {
      if(this.stateManager.appConfig.clickToPopup && isClick)
      {
        var tp = e.toolTip.items.items[0], at = tp.getActiveTab();
        if(at && at.title == 'MutantSeq')
          tp.setActiveTab(0); // we need to do this to make sure stats are updated
        e.toolTip.shouldShow = true;
        e.toolTip.show();
      }
      return;
    }
    var info = e.toolTipInfo;
    this.commentConf.width = info.type == 'row'? 800 : 436;
    this.commentConf.height = info.type == 'row'? 400 : 270;
    var t = new Ext.ToolTip({
        target: el,
        title: info.name,
        height: this.commentConf.height + 5,
        width: this.commentConf.width + 5,
        showDelay: (this.stateManager.appConfig.clickToPopup)? 0 : 500,
        shouldShow: true,
        autoHide: false,
        closable: true,
        draggable:true,
        listeners:{ show:this._hiliteTD, hide: this._loliteTD, beforeshow:this._checkClick }
    });
    e.toolTip = t;
    // display comments
    t.insert(1, new BMS.AG.ML.DynamicViewer.CommentManager(
    {
      info:info, data:info.com, viewer: this, td: el, conf: this.commentConf
    }));
    this.toolTips.push(t);
    t.targetXY = event.getXY(); // seems a bug, whenever one creates tooltip dynamically and tries to show it (without having tooltip show through its listening to mouse events), it fails with a undefined targetXY error
    t.showTimer = t.show.defer(t.showDelay, t); // only way to avoid toolTips pop up too fast (delayShow won't work)
  },
  _checkClick: function(e)
  {
    var tmp = e.shouldShow;
    e.shouldShow = false;
    return tmp;
  },
  _destroyToolTips: function()
  {
    if(this.toolTips)
      for(var i = 0; i < this.toolTips.length; i++)
        this.toolTips[i].destroy();
    this.toolTips = new Array();
  },
  getInfo: function(td)
  {
    var info = { isCell:true, td:td };
    info.tr = info.td.parentNode;
    info.type = info.tr.className;
    info.cellIndex = info.td.cellIndex; // not useful after using colspan
    info.cellSpanX = this._findCellSpanX(info.tr, info.cellIndex); // use this when using colspan
    info.rowIndex = info.tr.rowIndex;
    info.table = info.td.offsetParent;
    info.rows = info.table.rows;
    for(var i = 0; i < info.rows.length; i++)
      if(info.rows[i].className == info.type)
        break;
    info.baseRowIndex = i;
    info.baseCellIndex = this._findCellIdx4SpanX(info.rows[i], info.cellSpanX);
    info.baseRow = info.rows[i];
//     info.baseCell = info.baseRow.children[info.baseCellIndex]; // unused and gives strange errors when debugging (might be firebug problem)
    return info;
  },
  _getTargetInfo: function(e) // e = Ext.EventObject
  {
    var td = e.getTarget(), info;
    if(td.className == 'idcol') td = td.offsetParent; // I switched to use <div> inside <td> to restrict width for idcol, otherwise FF just doesn't respect width (despite CSS stipulation that table cell, block level elements width/max-width should be respected)
    if(td.nodeName == 'TD' && td.offsetParent && td.offsetParent.id  == this.id) //e.within has some issues
      return this.getInfo(td);
    else return { isCell:false, td:td };
  },
  // find the starting X for the given cellIndex by adding colspan together
  _findCellSpanX: function(tr, cellIndex)
  {
    var colspan = 0;
    for(var i = 0; i < cellIndex; i++)
      colspan += tr.children[i].colSpan;
    return colspan;
  },
  // find the corresponding cell idx at given row for the given cell
  _findCellIdx4SpanX: function(tr, cellSpanX)
  {
    var colspan = 0;
    for(var i = 0; i < tr.children.length; i++)
    {
      if(colspan == cellSpanX) return i;
      colspan += tr.children[i].colSpan;
    }
    return -1;
  },
  _controllerCallback: function(x, y)
  {
    if(x < 1) x = 1;
    if(y < 1) y = 1;
    this.stateManager.config.databox.xstart = x;
    this.stateManager.config.databox.ystart = y;
    this.calledByController = true;
    this.getDataDisplay();
    this.log('Action', 'Dragged view window');
  }
};
function replaceColor(e, className) // e is HTMLelement
{
  if(e.style.backgroundColor)
  {
    e.oldBGColor = e.style.backgroundColor;
    e.style.backgroundColor = '';
  }
  Ext.fly(e).addClass(className || 'lowlite');
}
function restoreColor(e, className)
{
  Ext.fly(e).removeClass(className || 'lowlite');
  if(e.oldBGColor && e.className.match(/^\s*$/)) e.style.backgroundColor = e.oldBGColor;
}

