////////////////////////////////////////////////////////////
// tracks should be of this JSON format:
// [
//   // note: a base track is required for every tclass, a base DNA is required for base protein if there's offset
//   // also base tracks should appear before any nonbase tracks
//   { // track 1
//    * id:12345, // real unique id
//     name:'sample1 protein'/'NM_12345', // descriptive name
//     group:'patient1', // grouping that this track belongs to
//     srcId: 12345, // protein-track only, denotes the id for its corresponding DNA track
//     desc:'this is a sample taken at 96 hrs', //optional
//     infoHtml: 'extra info:<b>test</b>', // optional, will be displayed at end of "Info" tab of row tooltip
//    * type:'base'/'feature'/'mutation', // base tracks always displays even when view is scrolled down, unless that tclass is hidden by user's choice or in "only mutation" mode, contains default values like color,offset,width for all same tclass tracks; mutation tracks should have a rate of mutation, feature tracks are simple overlay tracks
//    * tclass:'dna'/'protein'/'domain'/..., // type of this track
//     color:[255,0,0], // optional, the [r,g,b] color to display at 100% (mutations could have less than 100%, so actual color will be calculated based on mutation rate and this base color)
//     offset: 8, //required if CDS does not start at n.t. position 1
//    * seq: // base tracks should have only 1 seq!!
//     [{
//       id:'kinase domain', // optional, useful for special features (mutiple consecutive features with different colors sharing same id could solve displaying one domain whose a.a. are differentially highlighted due to mutation at some positions)
//       tx:138, // default is 1, required otherwise - the 1-based absolute base number in the entire alignment space (not the base/a.a. number on dna/protein sequence)
//       w:1, // width of this feature in its native unit (for domain it's a.a., not n.t.!),optional,default is 1
// note: either the b or s below is required for features that have sequences
//       b:[['A',3.5],['G',18.3]], // actual sequence/mutation rate when type is mutation
//       s:'ACGT--CGT'/'MSTPY', // actual sequence when type is base/feature, special char: ' ' no data, '-' gap, '?' unknown (color as no data), '.' muation reverted to non-mutation after user investigation
//       m:21.8, // optional,total mutational rate (this viewer calculates total as addition, use this to override)
//       c:[135,135,135] // optional color for this unit (it could be different than the whole track's color)
//     }]
//   }
// ]
// NOTE: frameshifting mutations will create problems for display!!
////////////////////////////////////////////////////////////

BMS.AG.ML.DynamicViewer.Track = function(data, parent)
{
  this.parent = parent;
  this.viewer = parent.parent;
  this.forceXtoN = true; // some software (sequencher) can't deal with X in DNA sequences
  Ext.apply(this, data);
  if(!this.name) this.name = this.id;
};
BMS.AG.ML.DynamicViewer.Track.prototype =
{
  // track should not be displayed
  noDisplay: function()
  {
    return !this.viewer.stateManager.config.displayTracks[this.tclass].show || this.failedFilter;
  },
  getHtml: function()
  {
    var tmpstr = new Array(), conf = this.viewer.stateManager.config;

    // get some settings
    var offset = this.getOpt('offset');
    var unitwidth = this.getOpt('width');
    var color = this.getOpt('color');

    // set base track color
    var rowcolor = (this.type == 'base' && color)? ' style="background-color:' + BMS.AG.ML.Util.getColor(color) + '"' : '';

    // add ID column
    tmpstr.push('<tr class="',this.tclass,'"',rowcolor,'><td class="idcol"><div class="idcol" title="',this.name,'">',this.name,'</div></td>');

    // padding column
    var startpad = (offset > conf.databox.xstart)? 1 : 0;
    if(startpad) tmpstr.push(this._colSpan(offset - conf.databox.xstart));

    if(this.type == 'base') // add base track HTML
    {
      // calculate track start/end
      var txstart, txend, seqstr = (this.seq && this.seq.length)? this.seq[0].s : '';
      txstart = (conf.databox.xstart > offset)? conf.databox.xstart : offset;
      txend = (conf.databox.xstart + conf.databox.width - 1 > offset)? conf.databox.xstart + conf.databox.width - 1 : offset;

      // record some useful info that'll be inherited by the non-base tracks of same tclass
      this.tstart = Math.floor((txstart - offset)/unitwidth) + 1;
      this.tend = Math.floor((txend - offset)/unitwidth) + 1;
      this.tend = (seqstr.length >= this.tend)? this.tend : seqstr.length;
      this.startpad = startpad;

      if(seqstr.length)
      {
        // add the sequence columns for the part of the sequence in current view window
        var seq = seqstr.substr(this.tstart - 1, this.tend - this.tstart + 1);
        if(this.tclass.match('coords'))
        {
          // fspan = any incomplete coords number + space (like '' in the '385', or '  ' in the '  385', or '84  ' in '84  385')
          var fspan = (seq.match(/^( +)/))? RegExp.$1.length : 0;
          if(!fspan && this.tstart-1 && seqstr.charAt(this.tstart-2) != ' ') //partial # at start of window
            fspan = (seq.match(/^(\S+ *)/))? RegExp.$1.length : seq.length;
          if(fspan < seq.length) // still has more space/num on the track in window
          {
            var colspan = fspan * unitwidth;
            if(conf.databox.xstart > offset && unitwidth > 1) // check frameshift
              colspan -= (conf.databox.xstart - offset) % unitwidth;
            // note: we don't display partial #s
            if(colspan > 0) tmpstr.push(this._colSpan(colspan));
            // now display the rest
            var first, seq = seq.substr(fspan);
            while(seq.match(/^(\S+ *)/))
            {
              var ts = RegExp.$1, seq = RegExp.rightContext, nspan = RegExp.$1.length * unitwidth;
              if(unitwidth > 1 && (first++ || colspan >=0))
              {
                var tspan = Math.floor(unitwidth/2);
                nspan -= tspan;
                tmpstr.push(this._colSpan(tspan));
              }
              tmpstr.push(this._colSpan(nspan, ts));
            }
//             for(var j = fspan; j < seq.length; j+=this.step)
//             {
//               var os = (j > fspan || colspan >=0)? Math.floor(unitwidth/2): 0,
//                   nspan = (seq.length >= j+this.step)? this.step : seq.length - j;
//               var tspan = nspan * unitwidth - os;
//               if(j == fspan && colspan < 0) tspan += colspan;
//               if(os) tmpstr.push('<td></td>');
//               if(nspan)
//                 tmpstr.push('<td' + ((nspan)? ' colspan='+tspan:'') + '>' + seq.substr(j, nspan) + '</td>');
//             }
          }
        }
        else
        {
          seq = seq.split('');
          for(var j = 0; j < seq.length; j++)
          {
            var colspan = unitwidth;
            if(j == 0)
            {
              if(conf.databox.xstart > offset && unitwidth > 1)
              {
                colspan = unitwidth - (conf.databox.xstart - offset) % unitwidth;
                this.firstColSpan = colspan;
              }
              else delete this.firstColSpan;
            }
            var colspanstr = (unitwidth > 1)? ' colspan=' + colspan : '';
            var classStr = (seq[j] == '-')? ' class="gap"' : ''; // special char
            tmpstr.push('<td',colspanstr,classStr,'>',((seq[j] && conf.size > 4)? seq[j] : ''),'</td>');
          }
        }
      }
    }
    else if(this.seq && this.seq.length) // now deal with the feature tracks
    {
      var totalColSpan = 0, baseTrack = this.base;
      this.featureInView = new Array();

      this.seq.sort(function(a,b){return sortSeq(a,b)});
      var lastx = baseTrack.tstart - 1, idx = startpad, first = true, seenx = {};
      for(var m = 0; m < this.seq.length; m++)
      {
        var x = this.seq[m].x || this.seq[m].tx;
        // skip feature unless it's in current view window
        if(x < baseTrack.tstart || x > baseTrack.tend || seenx[x]) continue; // seenx: guard against features reported twice by accident
        seenx[x] = 1;

        // sum up the mutation rate
        var sum = 100 - this.seq[m].orig[1], orig = this.seq[m].orig[0];
        this.seq[m].baseCellIndex = x - baseTrack.tstart + startpad + 1; // index of the base track cell that corresponds to this cell

        // calculate colspan for spacer TDs
        var colspan = (x - lastx - 1) * unitwidth;
        var spanAdjust = first && baseTrack.firstColSpan? unitwidth - baseTrack.firstColSpan : 0; // the first cell might need some adjustment if unitwidth > 1 (e.g., first a.a. cell is only 2 nt wide instead of 3)
        if(first) colspan -= spanAdjust;
        lastx = x;
        var spacercell = '', featColSpan = unitwidth;
        // deal with colspan issue - IE and FF has 1000 as colspan limit (not W3C spec and Chrome does not have the limit, although Chrome's slower than even FF at rendering large tables)
        if(colspan > 0)
        {
          if(colspan > 1000)
          {
            var t = this._splitColspan(colspan);
            spacercell = t.join('');
            idx += t.length;
          }
          else
          {
            spacercell = this._colSpan(colspan);
            idx++; // for spacer
          }
        }
        else if(first) featColSpan -= spanAdjust;
        first = false;
        idx++; // for feature
        var colspanstr = (featColSpan > 1)? ' colspan="' + featColSpan + '"' : '';
        totalColSpan += colspan + featColSpan;
        var ihtml = ((conf.size > 6)? ((this.seq[m].b.length > 1)? 'X' :
                  ((this.seq[m].b.length > 0)? this.seq[m].b[0][0] : '')) : '');// when silent mutation, b.length = 0
        // special char
        var bgcolor = (ihtml == '-')? ' class="gap"' : (ihtml.match(/[ ?]/))? ' class="nodata"' : (ihtml == '.')? ' class="removed"' : ' style="background-color:' + this._getMutColor(sum) + ';"';
        tmpstr.push(spacercell,'<td' , colspanstr , bgcolor,'>' , ihtml, '</td>');
        if(ihtml != ' ') this.featureInView.push([this.seq[m], idx]);
      }
      if(totalColSpan < conf.databox.width)
        tmpstr.push('<td colspan=0></td>'); // colspan=0 is poorly supported but this td is needed for FF if one uses border on the last td in row (the border bleeds over)
    }
    tmpstr.push('</tr>\n');
    return tmpstr.join('');
  },
  //
  reprocess: function()
  {
    this.offset = this.getOpt('offset') || 1;
    if(!this.viewer.stateManager.config.displayTracks.dna.show)
      this.offset = 1; // offset only useful for protein/dna, and if dna not displayed, no need for protein offset
  },
  _replaceX: function(seq)
  {
    if(this.tclass == 'dna' && this.forceXtoN) seq = seq.replace(/X/g, 'N');
    return seq;
  },
  getAlignSeq: function(rate) // with gaps
  {
    this.assembleSeq(rate);
    if(this.realSeq && this.realSeq.length)
      return this._replaceX(this.realSeq[0].s);
  },
  getSeq: function(rate) // does not contain gaps
  {
    this.assembleSeq(rate);
    if(this.realSeq && this.realSeq.length)
      return this._replaceX(this.realSeq[0].s.replace(/[- ]+/g,'')); // special char
  },
  getBaseSeqAt: function(x, isGapped)
  {
    return this.base.seq[0].s.charAt((isGapped? x : this.addGap(x)) - 1);
  },
  getSeqAt: function(x, isGapped)
  {
    if(this.realSeq && this.realSeq.length)
      return this.realSeq[0].s.charAt((isGapped? x : this.addGap(x)) - 1);
  },
  getFeatureAt: function(x, isGapped)
  {
    var feat, gx = isGapped? x : this.addGap(x), ugx = isGapped? this.removeGap(x) : x;
    if(this.seq && this.seq.length)
    {
      for(var i = 0; i < this.seq.length; i++)
      {
        if(this.seq[i].tx == ugx)
        {
          feat = this.seq[i];
          break;
        }
      }
    }
    return { base: this.getBaseSeqAt(x, isGapped),
             seq: this.getSeqAt(x, isGapped),
             feat: feat,
             pos: ugx };
  },
  // x1, x2 are gapped (gets only the sequence b/w x1 and x2 NOT INCLUDING x1 and x2)
  getBaseGSeqBtn: function(x1, x2)
  {
    return x2? this.base.seq[0].s.substr(x1, x2 - x1 - 1) : this.base.seq[0].s.substr(x1);
  },
  // process all features and set original & its percent
  setOrig: function()
  {
    if(this.seq)
    {
      for(var m = 0; m < this.seq.length; m++)
      {
        var x = this.seq[m].x || this.seq[m].tx;
        // sum up the mutation rate
        var sum = 0, orig = this.getBaseSeqAt(x, true), mut = []; // tx Charles sent in is actually gapped
        for(var n = 0; n < this.seq[m].b.length; n++)
          if(this.seq[m].b[n][0] != orig)
          {
            sum += this.seq[m].b[n][1];
            mut.push(this.seq[m].b[n]);
          }
        this.seq[m].b = mut; // make sure b array doesn't have original
        this.seq[m].orig = [orig, 100-sum];
      }
    }
  },
  // converting gapped X to non-gapped X
  removeGap: function(x) // formally _adjustX
  {
    var t = this.dataGaps, rx = x;
    if(!t || !t.length) return x;
    for(var m = 0; m < t.length; m++)
    {
      if(t[m][0] > x)
      {
        if(m > 0)
        {
          if(t[m][0] - (t[m][1] - t[m-1][1]) <= x) // x fell inside a gap
            return t[m][0] - t[m][1];
          else // inside previous non-gap
            return x - t[m-1][1];
        }
        else if(t[m][0] - t[m][1] <= x) return t[m][0] - t[m][1];// x fell inside the 1st gap
        else return x; // before any gap
      }
    }
    return x - t[t.length-1][1]; // after all gaps
  },
  // converting non-gapped X to gapped X
  addGap: function(x)
  {
    var t = this.dataGaps, rx = x;
    if(!t || !t.length) return x;
    for(var m = 0; m < t.length; m++)
    {
      if(t[m][0] - t[m][1] > x)
      {
        if(m > 0) return x + t[m-1][1];
        else return x; // before any gap
      }
    }
    return x + t[t.length-1][1]; // after all gaps
  },
  _guessWidth: function(tclass)
  {
    switch(tclass)
    {
      case 'prot':
      case 'domain':
      case 'protcoords': return 3;
      default: return 1;
    }
  },
  // colors mutations based on its mutation rate
  _getMutColor: function(sum, color)
  {
    if(!color || color.length < 3) color = [255, 200, 200];
    var ratio = (100 - sum) / 100;
    return BMS.AG.ML.Util.getColor([color[0], Math.floor(color[1] * ratio) + 55, Math.floor(color[2] * ratio) + 55]);
  },
  // concatenate all mutations in the same feature location
  _mutStr: function(b)
  {
    var str = new Array();
    b.sort(function(a,b){return a[0]>b[0] ? 1 : a[0]<b[0] ? -1 : 0});
    for(var j = 0; j < b.length; j++)
      str.push(b[j][0]);
    return str.join(';');
  },
  _splitColspan: function(colspan)
  {
    var t = new Array();
    for(var i = 0; i < colspan; i += 1000)
    {
      if(i <= colspan - 1000) t.push('<td colspan=1000></td>');
      else t.push('<td colspan='+ (colspan - i) +'></td>');
    }
    return t;
  },
  _colSpan: function(colspan, str)
  {
    return '<td' + ((colspan > 1)? ' colspan=' + colspan:'') + '>'+ (str? str:'') +'</td>';
  },
  // get an option based on current track value or default value
  getOpt: function(attr)
  {
    var a = this[attr] || this.base[attr], conf = this.viewer.stateManager.config;
    switch(attr)
    {
      case 'width': return (a || this._guessWidth(this.tclass)) / this.parent.widthLCF;
      case 'offset': return conf.displayTracks.dna.show? a || 1 : 1; // offset only useful for protein/dna, and if dna not displayed, no need for protein offset
      default: return a;
    }
  },
  // assemble the current sequence using base seq
  assembleSeq: function(thresh) // default to 20 (%)
  {
    if(thresh === undefined) thresh = 20;
    if(this.realSeq && this.mutThresh === thresh) return;
    this.mutThresh = thresh;
    var z = new Array();
    var startx = 0;
    if(!this.seq || !this.seq.length) return;
    this.realSeq = [];
    if(this.type == 'base') { this.realSeq[0] = this.seq[0]; return; }
    this.seq.sort(function(a,b){return sortSeq(a,b)});
    // note: we assume that any gap on the track is provided to us in this.seq[m].s!! for .b mutations, we assume backend provides tx correctly
    for(var m = 0; m < this.seq.length; m++)
    {
      var x = this.seq[m].x || this.seq[m].tx;
      z.push(this.getBaseGSeqBtn(startx, x));
      startx = x;
      var tmp = this.seq[m].b, orig = this.seq[m].orig;
      
      if(tmp && tmp.length) // mutation // special char
      {
        // note: orig is NOT in tmp array!!!!
        if(tmp[0][0] == '.') z.push(orig[0]); // user removed the mutation
        else if(/[? -]/.test(tmp[0][0])) z.push(tmp[0][0]); // unsure (like missing one nt in a triplet), no data, gap
        else 
        {
          var good = [];
          // set threshold
          if(this.seq[m].orig[1] >= thresh) good.push(orig[0]);
          for(var i = 0; i < tmp.length; i++)
            if(tmp[i][1] >= thresh) good.push(tmp[i][0]);
          if(this.tclass == 'prot') z.push(good.length > 1? 'X':good[0]);
          else // map nucleotides for DNA
          {
            var code = {CT:'Y',AG:'R',AT:'W',CG:'S',GT:'K',AC:'M',AGT:'D',ACG:'V',ACT:'H',CGT:'B',ACGT:'N'};
            var bases = good.sort().join(''), mapped = code[bases]; // do I need toUpperCase?
            if(!mapped && /^[ACGTUXYRWSKMDVHBN?]$/.test(bases)) mapped = bases;
            if(!mapped && MutationViewer.alertNoCode) BMS.AG.ML.Util.alert('Error!!', 'No code found for ' + bases + ' at position ' + x + ' of ' + this.name);
          z.push(mapped || '?');
        }
      }
    }
      else z.push(orig[0]); // this probably doesn't happen
    }
    z.push(this.getBaseGSeqBtn(startx));
    this.realSeq.push({s:z.join(''),x:1});
  },
  // assemble a special-format sequence
  assembleMutSeq: function(thresh) // default to 20 (%)
  {
    if(thresh === undefined) thresh = 20;
    var z = [];
    var startx = 0;
    if(!this.seq || !this.seq.length) return;
    if(this.type == 'base')
      return this._replaceX(this.seq[0].s.replace(/- /g,''));
    else
    {
      var max = 1; // max length of mutation array (corresponds to max height of formatted seq)
      this.seq.sort(function(a,b){return sortSeq(a,b)});
      // note: we assume that any gap on the track is provided to us in this.seq[m].s!! for .b mutations, we assume backend provides tx correctly
      for(var m = 0; m < this.seq.length; m++)
      {
        var x = this.seq[m].x || this.seq[m].tx, y = this.getBaseGSeqBtn(startx, x).replace(/[ -]+/g, '').split('');
        for(var i = 0; i < y.length; i++)
          z.push(y[i]); // push in all non-mutations before current mutation
        startx = x;
        var tmp = this.seq[m].b, orig = this.seq[m].orig;

        if(tmp && tmp.length) // mutation // special char
        {
          // note: orig is NOT in tmp array!!!!
          if(tmp[0][0] == '.')
          {
            if(!/[ -]/.test(orig[0])) z.push([orig[0]]); // user removed the mutation
          }
          else if(/[ -]/.test(tmp[0][0])) ; // skip no data, gap
          else if(tmp[0][0] == '?') z.push(['?']); // unsure (like missing one nt in a triplet)
          else
          {
            var good = [], f = [];
            if(this.seq[m].orig[1] >= thresh) good.push(orig);
            for(var i = 0; i < tmp.length; i++)
              if(tmp[i][1] >= thresh) good.push(tmp[i]);
            good.sort(function(a,b){return b[1]-a[1]});
            for(var i = 0; i < good.length; i++)
              f.push(good[i][0]);
            if(f.length > max) max = f.length;
            z.push(f);
          }
        }
        else z.push([orig[0]]); // this probably doesn't happen
      }
      var y = this.getBaseGSeqBtn(startx).replace(/[ -]+/g, '').split('');
      for(var i = 0; i < y.length; i++)
        z.push(y[i]); // push in all non-mutations before current mutation
      // now assemble the sequence
      var arr = [];
      for(var i = 0; i <= max; i++)
        arr[i] = [];
      for(var i = 0; i < z.length; i++)
      {
        // set coord track
        if(!((i-1) %10))
        {
          var num = i-1? ((i-1) + '').split('') : [], j = 0, k = i-1? 10:9;
          for(; j < num.length; j++)
            arr[0].push(num[j]);
          for(; j < k; j++)
            arr[0].push(j == k-5? '.' : ' ');
        }

        // work on other tracks now
        // the following way displays like:
        //    L
        // YSPVHAK
        //    IWP
//         var t = z[i], l = t.length, j = 1, k = 0;
//         for(; j <= Math.floor((max - l)/2); j++, k++)
//           arr[j].push(' ');
//         for(k = 0; k < l; k++)
//           arr[j+k].push(t[k]);
//         for(k = j + l; k <= max; k++)
//           arr[k].push(' ');
          
        // the following way displays like:
        // YSPLHAK
        //    VWP
        //    I
        var t = z[i], l = t.length, k = 0;
        for(; k < l; k++)
          arr[k+1].push(t[k]);
        for(; k < max; k++)
          arr[k+1].push(' ');
      }
      for(var i = 0; i <= max; i++)
        arr[i] = arr[i].join('');
      return this._replaceX(arr.join('\n'));
    }
  },
  // record a sorted list of gap positions on this track
  // gap positions stored as [nextNonGapPosition, totalGapSizeSoFar]
  setGaps: function()
  {
    var z = new Array(), g = 0;
    if(!this.realSeq || !this.realSeq.length || this.dataGaps) return;
    var basex = this.realSeq[0].x || (this.realSeq[0].tx + g) || 1;
    var idx = basex, s = this.realSeq[0].s;
    while(s.match(/([- ]+)/)) // special char
    {
      var gapSize = RegExp.$1.length;
      idx += RegExp.leftContext.length + gapSize;
      g += gapSize;
      s = RegExp.rightContext;
      z.push([idx, g]);
    }
    this.dataGaps = z;
  },
  // is this track a baseline track for a group?
  isBL: function()
  {
    if(this.ClinicalTimePoint) return this.ClinicalTimePoint == 'BL';
    var com = this.viewer.comments;
    if(com && com.row && com.row[this.id])
    {
      for(var i = 0; i < com.row[this.id].length; i++)
        if(com.row[this.id][i][0] == 'ClinicalTimePoint')
          return com.row[this.id][i][1] == 'BL';
    }
    return false;
  }
};
function sortSeq(a, b)
{
  var ax = (a.x)? a.x : a.tx;
  var bx = (b.x)? b.x : b.tx;
  return ax - bx;
}